from __future__ import absolute_import, division, print_function, with_statement

from os import path as p
from functools import partial
from multiprocessing.dummy import Pool as ThreadPool
from multiprocessing import Pool
from time import time, sleep

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.task import react

from riko.lib.collections import (
    SyncPipe, SyncCollection, get_chunksize, get_worker_cnt)

from riko.twisted.collections import AsyncPipe, AsyncCollection
from riko.twisted.utils import asyncImap, asyncSleep
from riko.modules.pipefetch import pipe, asyncPipe

NUMBER = 3
LOOPS = 3
DELAY = 0.2

parent = p.join(p.abspath(p.dirname(p.dirname(__file__))), 'data')
files = [
    'blog.ouseful.info_feed.xml',
    'feed.xml',
    'feeds.delicious.com_v2_rss_popular?count=3.xml',
    'feeds.delicious.com_v2_rss_popular?count=15.xml',
    'feeds.delicious.compsychemedia.xml',
    'feeds.feedburner.com_ouseful.xml',
    'feeds.feedburner.com_TheEdTechie.xml',
    'feeds.feedburner.com_yodelanecdotal.xml',
    'feeds.gawker.com_jalopnik_full.xml',
    'news.yahoo.com_rss_health.xml',
    'news.yahoo.com_rss_topstories.xml',
    'www.autoblog.com_rss.xml',
    'www.fourtitude.com_news_publish_rss.xml',
    'www.greenhughes.com_rssfeed.xml',
    'www.slideshare.net_rss_user_psychemedia.xml']

get_path = lambda name: 'file://%s' % p.join(parent, name)
sources = [{'url': get_path(f)} for f in files]
length = len(files)
conf = {'url': [{'value': get_path(f)} for f in files], 'sleep': DELAY}
iterable = [DELAY for x in files]


def baseline_sync():
    return map(sleep, iterable)


def baseline_threads():
    workers = get_worker_cnt(length)
    chunksize = get_chunksize(length, workers)
    pool = ThreadPool(workers)
    return list(pool.imap_unordered(sleep, iterable, chunksize=chunksize))


def baseline_procs():
    workers = get_worker_cnt(length, False)
    chunksize = get_chunksize(length, workers)
    pool = Pool(workers)
    return list(pool.imap_unordered(sleep, iterable, chunksize=chunksize))


def sync_pipeline():
    return list(pipe(conf=conf))


def sync_pipe():
    return SyncPipe('fetch', conf=conf).list


def sync_collection():
    return SyncCollection(sources, sleep=DELAY).list


def par_sync_collection():
    return SyncCollection(sources, parallel=True, sleep=DELAY).list


def baseline_async():
    return asyncImap(asyncSleep, iterable)


def async_pipeline():
    d = asyncPipe(conf=conf)
    d.addCallbacks(list, print)
    return d


def async_pipe():
    return AsyncPipe('fetch', conf=conf).list


def async_collection():
    return AsyncCollection(sources, sleep=DELAY).list


def parse_results(results):
    switch = {0: 'secs', 3: 'msecs', 6: 'usecs'}
    best = min(results)

    for places in [0, 3, 6]:
        factor = pow(10, places)
        if 1 / best // factor == 0:
            break

    return round(best * factor, 2), switch[places]


def print_time(test, max_chars, run_time, units):
    padded = test.zfill(max_chars).replace('0', ' ')
    msg = '%s - %i repetitions/loop, best of %i loops: %s %s'
    print(msg % (padded, NUMBER, LOOPS, run_time, units))


@inlineCallbacks
def run_async(reactor, tests, max_chars):
    for test in tests:
        results = []

        for i in range(LOOPS):
            loop = 0

            for j in range(NUMBER):
                start = time()
                yield test()
                loop += time() - start

            results.append(loop)

        run_time, units = parse_results(results)
        print_time(test.func_name, max_chars, run_time, units)

    returnValue(None)

if __name__ == '__main__':
    from timeit import repeat

    run = partial(repeat, repeat=LOOPS, number=NUMBER)
    sync_tests = [
        'baseline_sync', 'baseline_threads', 'baseline_procs', 'sync_pipeline',
        'sync_pipe', 'sync_collection', 'par_sync_collection']

    async_tests = [baseline_async, async_pipeline, async_pipe, async_collection]
    combined_tests = sync_tests + [f.func_name for f in async_tests]
    max_chars = max(map(len, combined_tests))

    for test in sync_tests:
        results = run('%s()' % test, setup='from __main__ import %s' % test)
        run_time, units = parse_results(results)
        print_time(test, max_chars, run_time, units)

    react(run_async, [async_tests, max_chars])



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.lib.log
~~~~~~~~~~~~
Provides a logging class

"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import sys
import logging

from builtins import *


class LogFilter(logging.Filter):
    """Filters (lets through) all messages with level < LEVEL"""
    # http://stackoverflow.com/a/24956305/408556
    def __init__(self, level):
        self.level = level

    def filter(self, record):
        return record.levelno < self.level


class Logger(object):
    """
    # logger = Logger('name', 'DEBUG').logger
    # logger.debug("A DEBUG message")
    # logger.info("An INFO message")
    # logger.warning("A WARNING message")
    # logger.error("An ERROR message")
    # logger.critical("A CRITICAL message")
    """
    def __init__(self, name, level='INFO', context=None):
        level = 'DEBUG' if context and context.verbose else level
        self.level = getattr(logging, level)
        self.name = name

    @property
    def logger(self):
        # messages < level go to stdout
        # messages >= level (and >= logging.WARNING) go to stderr
        stdout_hdlr = logging.StreamHandler(sys.stdout)
        stderr_hdlr = logging.StreamHandler(sys.stderr)
        log_filter = LogFilter(self.level)
        stdout_hdlr.addFilter(log_filter)
        stdout_hdlr.setLevel(self.level)
        stderr_hdlr.setLevel(max(self.level, logging.WARNING))

        logger = logging.getLogger(self.name)
        logger.addHandler(stdout_hdlr)
        logger.addHandler(stderr_hdlr)
        return logger



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipefetchpage
~~~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for fetching web pages.

Fetches the source of a given web site as a string. This data can then be
converted into an RSS feed or merged with other data in your Pipe using the
`regex` module.

Examples:
    basic usage::

        >>> from riko.modules.pipefetchpage import pipe
        >>> from . import FILES
        >>>
        >>> conf = {'url': FILES[5], 'start': '<title>', 'end': '</title>'}
        >>> next(pipe(conf=conf))['content']  # doctest: +ELLIPSIS
        u'CNN.com International - Breaking, World..., Entertainment and...'

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from builtins import *
from six.moves.urllib.request import urlopen
from twisted.internet.defer import inlineCallbacks, returnValue

from . import processor
from riko.lib import utils
from riko.lib.log import Logger
from riko.lib.tags import get_text
from riko.twisted import utils as tu

OPTS = {'ftype': 'none'}
logger = Logger(__name__).logger


def get_string(content, start, end):
    # TODO: convert relative links to absolute
    # TODO: remove the closing tag if using an HTML tag stripped of HTML tags
    # TODO: clean html with Tidy
    content = content.decode('utf-8')
    start_location = content.find(start) if start else 0
    right = content[start_location + len(start):]
    end_location = right[1:].find(end) + 1 if end else len(right)
    return right[:end_location] if end_location > 0 else right


@inlineCallbacks
def asyncParser(_, objconf, skip, **kwargs):
    """ Asynchronously parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: content)
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>> from riko.lib.utils import Objectify
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x[0])['content'][:32])
        ...     conf = {'url': FILES[5], 'start': '<title>', 'end': '</title>'}
        ...     objconf = Objectify(conf)
        ...     kwargs = {'stream': {}, 'assign': 'content'}
        ...     d = asyncParser(None, objconf, False, **kwargs)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        CNN.com International - Breaking
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        content = yield tu.urlRead(url)
        parsed = get_string(content, objconf.start, objconf.end)
        detagged = get_text(parsed) if objconf.detag else parsed
        splits = detagged.split(objconf.token) if objconf.token else [detagged]
        stream = ({kwargs['assign']: chunk} for chunk in splits)

    result = (stream, skip)
    returnValue(result)


def parser(_, objconf, skip, **kwargs):
    """ Parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from riko.lib.utils import Objectify
        >>> from . import FILES
        >>>
        >>> conf = {'url': FILES[5], 'start': '<title>', 'end': '</title>'}
        >>> objconf = Objectify(conf)
        >>> kwargs = {'stream': {}, 'assign': 'content'}
        >>> result, skip = parser(None, objconf, False, **kwargs)
        >>> next(result)['content'][:32]
        u'CNN.com International - Breaking'
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        content = urlopen(url).read()
        parsed = get_string(content, objconf.start, objconf.end)
        detagged = get_text(parsed) if objconf.detag else parsed
        splits = detagged.split(objconf.token) if objconf.token else [detagged]
        stream = ({kwargs['assign']: chunk} for chunk in splits)

    return stream, skip


@processor(async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A source that asynchronously fetches the content of a given web site as
    a string.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'start', 'end', 'token', or 'detag'.

            url (str): The web site to fetch
            start (str): The starting string to fetch (exclusive, default:
                None).

            end (str): The ending string to fetch (exclusive, default: None).
            token (str): The tokenizer delimiter string (default: None).
            detag (bool): Remove html tags from content (default: False).

        assign (str): Attribute to assign parsed content (default: content)

    Returns:
        dict: twisted.internet.defer.Deferred item

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x))
        ...     path = 'value.items'
        ...     conf = {'url': FILES[4], 'start': 'DOCTYPE ', 'end': 'http'}
        ...     d = asyncPipe(conf=conf)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        {u'content': u'html PUBLIC "-//W3C//DTD XHTML+RDFa 1.0//EN" "'}
    """
    return asyncParser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A source that fetches the content of a given web site as a string.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'start', 'end', 'token', or 'detag'.

            url (str): The web site to fetch
            start (str): The starting string to fetch (exclusive, default:
                None).

            end (str): The ending string to fetch (exclusive, default: None).
            token (str): The tokenizer delimiter string (default: None).
            detag (bool): Remove html tags from content (default: False).

        assign (str): Attribute to assign parsed content (default: content)

    Yields:
        dict: item

    Examples:
        >>> from . import FILES
        >>> conf = {'url': FILES[4], 'start': 'DOCTYPE ', 'end': 'http'}
        >>> next(pipe(conf=conf))
        {u'content': u'html PUBLIC "-//W3C//DTD XHTML+RDFa 1.0//EN" "'}
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.lib.autorss
~~~~~~~~~~~~~~~~
Provides functions for finding RSS feeds from a site's LINK tags
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from itertools import chain
from html.parser import HTMLParser

from builtins import *
from six.moves.urllib.request import urlopen
from twisted.internet.defer import inlineCallbacks, returnValue

from riko.twisted.utils import urlOpen

TIMEOUT = 10


class LinkParser(HTMLParser):
    def reset(self):
        HTMLParser.reset(self)
        self.entry = []

    def handle_starttag(self, tag, attrs):
        entry = dict(attrs)
        alternate = entry.get('rel') == 'alternate'
        rss = 'rss' in entry.get('type', '')

        if (alternate or rss) and 'href' in entry:
            entry['link'] = entry['href']
            entry['tag'] = tag
            self.entry = chain(self.entry, [entry])

from riko.lib.log import Logger
logger = Logger(__name__).logger


def gen_entries(f, parser):
    for line in f:
        parser.feed(line)

        for entry in parser.entry:
            yield entry


@inlineCallbacks
def asyncGetRSS(url):
    # TODO: implement via an async parser
    # maybe get twisted.web.microdom.parse working for HTML
    parser = LinkParser()

    try:
        f = yield urlOpen(url, timeout=TIMEOUT)
    except ValueError:
        f = filter(None, url.splitlines())

    returnValue(gen_entries(f, parser))


def get_rss(url):
    parser = LinkParser()

    try:
        f = urlopen(url, timeout=TIMEOUT)
    except ValueError:
        f = filter(None, url.splitlines())

    return gen_entries(f, parser)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipeslugify
~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for slugifying text.

Examples:
    basic usage::

        >>> from riko.modules.pipeslugify import pipe
        >>> next(pipe({'content': 'Hello World'}))['slugify']
        'hello-world'

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from builtins import *
from slugify import slugify

from . import processor
from riko.lib.log import Logger

OPTS = {'ftype': 'text', 'ptype': 'none', 'field': 'content'}
DEFAULTS = {}
logger = Logger(__name__).logger


def parser(word, _, skip, **kwargs):
    """ Parsers the pipe content

    Args:
        item (obj): The entry to process (a DotDict instance)
        _ (None): Ignored.
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: exchangerate)
        stream (dict): The original item

    Returns:
        Tuple(dict, bool): Tuple of (item, skip)

    Examples:
        >>> from riko.lib.utils import Objectify
        >>>
        >>> item = {'content': 'Hello World'}
        >>> kwargs = {'stream': item}
        >>> parser(item['content'], None, False, **kwargs)[0]
        'hello-world'
    """
    parsed = kwargs['stream'] if skip else slugify(word.strip())
    return parsed, skip


@processor(DEFAULTS, async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A processor module that asynchronously slugifies the field of an item.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        assign (str): Attribute to assign parsed content (default: simplemath)
        field (str): Item attribute from which to obtain the first number to
            operate on (default: 'content')

    Returns:
        Deferred: twisted.internet.defer.Deferred item with concatenated content

    Examples:
        >>> from twisted.internet.task import react
        >>> from riko.twisted import utils as tu
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['slugify'])
        ...     d = asyncPipe({'content': 'Hello World'})
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        hello-world
    """
    return parser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A processor that slugifies the field of an item.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        assign (str): Attribute to assign parsed content (default: slugify)
        field (str): Item attribute from which to obtain the text to slugify
            (default: 'content')

    Yields:
        dict: an item with slugified content

    Examples:
        >>> next(pipe({'content': 'Hello World'}))['slugify']
        'hello-world'
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipexpathfetchpage
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for fetching the source of a given web site as DOM nodes or a
string.

This module fetches the source of a given web site as DOM nodes or a string.
This data can then be converted into a RSS/JSON feed or merged with other data
in your Pipe using the Regex module, String Builder modules and others that
will help achieve what you want.

By default, the module will output the DOM elements as items. You can
optionally use the "Emit items as string" checkbox if you need the html as a
string.

You can use the "Extract using XPATH" field to fine tune what you need from the
HTML Page. For example, if I want all the links in the page I can simply use
"/a" to grab all links. If I want all the images in the html I can do "/img".
Read more on XPATH. You can also find XPATH statements using firebug to target
data that you want in a HTML page.

You have the option to run the parser using support for HTML4 (by default) or
checking the "Use HTML5 parser" checkbox to use the HTML5 parser.

Examples:
    basic usage::

        >>> from . import FILES
        >>> from riko.modules.pipexpathfetchpage import pipe
        >>>
        >>> conf = {'url': FILES[1], 'xpath': '/rss/channel/item'}
        >>> title = 'Running “Native” Data Wrangling Applications'
        >>> next(pipe(conf=conf))['title'][:44] == title
        True

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from os.path import splitext

from builtins import *
from six.moves.urllib.request import urlopen
from lxml import objectify, html
from lxml.html import html5parser
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web import microdom

from . import processor
from riko.lib import utils
from riko.lib.log import Logger
from riko.twisted import utils as tu

OPTS = {'ftype': 'none'}
logger = Logger(__name__).logger


# TODO: convert relative links to absolute
# TODO: remove the closing tag if using an HTML tag stripped of HTML tags
# TODO: clean html with Tidy
def genXPATH(tree, xpath, pos=0):
    tags = xpath.split('/')[1:] if xpath else []
    elements = tree.getElementsByTagName(tags[pos]) if tags else [tree]

    if len(tags or [1]) - pos == 1:
        for element in elements:
            yield element
    else:
        for element in elements:
            for e in genXPATH(element, xpath, pos + 1):
                yield e


@inlineCallbacks
def asyncParser(_, objconf, skip, **kwargs):
    """ Asynchronously parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: content)
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>> from riko.lib.utils import Objectify
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x[0])['title'][:44])
        ...     xpath = '/rss/channel/item'
        ...     objconf = Objectify({'url': FILES[1], 'xpath': xpath})
        ...     kwargs = {'stream': {}}
        ...     d = asyncParser(None, objconf, False, **kwargs)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        Running “Native” Data Wrangling Applications
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        ext = splitext(url)[1].lstrip('.')
        f = yield tu.urlOpen(url)

        if ext == 'xml':
            root = yield microdom.parse(f)
            elements = genXPATH(root, objconf.xpath)
            items = map(tu.elementToDict, elements)
        elif ext == 'html':
            tree = html5parser.parse(f) if objconf.html5 else html.parse(f)
            root = tree.getroot()
            elements = root.xpath(objconf.xpath)
            items = map(utils.etree_to_dict, elements)
        else:
            raise TypeError('Invalid file type %s' % ext)

        stringified = ({kwargs['assign']: str(i)} for i in items)
        stream = stringified if objconf.stringify else items

    result = (stream, skip)
    returnValue(result)


def parser(_, objconf, skip, **kwargs):
    """ Parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from riko.lib.utils import Objectify
        >>> from . import FILES
        >>>
        >>> objconf = Objectify({'url': FILES[1], 'xpath': '/rss/channel/item'})
        >>> kwargs = {'stream': {}}
        >>> result, skip = parser(None, objconf, False, **kwargs)
        >>> title = 'Running “Native” Data Wrangling Applications'
        >>> next(result)['title'][:44] == title
        True
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        ext = splitext(url)[1].lstrip('.')
        f = urlopen(url)

        if ext == 'xml':
            tree = objectify.parse(f)
        elif ext == 'html':
            tree = html5parser.parse(f) if objconf.html5 else html.parse(f)

        root = tree.getroot()
        elements = root.xpath(objconf.xpath)
        items = map(utils.etree_to_dict, elements)
        stringified = ({kwargs['assign']: str(i)} for i in items)
        stream = stringified if objconf.stringify else items

    return stream, skip


@processor(async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A source that asynchronously fetches the content of a given website as
    DOM nodes or a string.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'xpath', 'html5', 'stringify', or 'assign'.

            url (str): The web site to fetch
            xpath (str): The XPATH to extract (default: None, i.e., return
                entire page)

            html5 (bool): Use the HTML5 parser (default: False)
            stringify (bool): Return the web site as a string (default: False)
            assign (str): Attribute to assign parsed content (default: content)

    Returns:
        dict: twisted.internet.defer.Deferred item

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['guid']['content'])
        ...     conf = {'url': FILES[1], 'xpath': '/rss/channel/item'}
        ...     d = asyncPipe(conf=conf)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        http://blog.ouseful.info/?p=12065
    """
    return asyncParser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A source that fetches the content of a given website as DOM nodes or a
    string.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'xpath', 'html5', 'stringify', or 'assign'.

            url (str): The web site to fetch
            xpath (str): The XPATH to extract (default: None, i.e., return
                entire page)

            html5 (bool): Use the HTML5 parser (default: False)
            stringify (bool): Return the web site as a string (default: False)
            assign (str): Attribute to assign parsed content (default: content)

    Yields:
        dict: item

    Examples:
        >>> from . import FILES
        >>> conf = {'url': FILES[1], 'xpath': '/rss/channel/item'}
        >>> next(pipe(conf=conf))['guid']['content']
        'http://blog.ouseful.info/?p=12065'
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
riko demo
~~~~~~~~~

Word Count

    >>> from riko.lib.collections import SyncPipe
    >>> from riko import get_path
    >>>
    >>> url = get_path('users.jyu.fi.html')
    >>> fetch_conf = {
    ...     'url': url, 'start': '<body>', 'end': '</body>', 'detag': True}
    >>> replace_conf = {'rule': {'find': '\\n', 'replace': ' '}}
    >>>
    >>> counts = (SyncPipe('fetchpage', conf=fetch_conf)
    ...     .strreplace(conf=replace_conf, assign='content')
    ...     .stringtokenizer(conf={'delimiter': ' '}, emit=True)
    ...     .count()
    ...     .output)
    >>>
    >>> next(counts)
    {u'count': 70}

Fetching feeds

    >>> from riko.modules.pipefetch import pipe as fetch
    >>>
    >>> url = get_path('feeds.gawker.com_jalopnik_full.xml')
    >>> feed = fetch(conf={'url': url})
    >>> item = next(feed)
    >>> sorted(item.keys())
    ['author', u'author.name', u'author.uri', 'content', u'dc:creator', 'id', \
'link', u'pubDate', 'summary', 'title', 'updated', 'updated_parsed', u'y:id', \
u'y:published', u'y:title']
    >>> item['title'], item['link']  # doctest: +ELLIPSIS
    (u'...', u'http...')
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from twisted.internet.defer import inlineCallbacks
from riko import get_path
from riko.lib.collections import SyncPipe
from riko.twisted.collections import AsyncPipe

replace_conf = {'rule': {'find': '\n', 'replace': ' '}}
url1 = get_path('news.yahoo.com_rss_health.xml')
url2 = get_path('www.caltrain.com_Fares_farechart.html')
fetch_conf = {
    'url': url2, 'start': '<body>', 'end': '</body>', 'detag': True}


def pipe(test=False):
    s1 = SyncPipe('fetch', test=test, conf={'url': url1}).output
    s2 = (SyncPipe('fetchpage', test=test, conf=fetch_conf)
        .strreplace(conf=replace_conf, assign='content')
        .stringtokenizer(conf={'delimiter': ' '}, emit=True)
        .count()
        .output)

    print(next(s1)['title'], next(s2)['count'])


@inlineCallbacks
def asyncPipe(reactor, test=False):
    s1 = yield AsyncPipe('fetch', test=test, conf={'url': url1}).output
    s2 = yield (AsyncPipe('fetchpage', test=test, conf=fetch_conf)
        .strreplace(conf=replace_conf, assign='content')
        .stringtokenizer(conf={'delimiter': ' '}, emit=True)
        .count()
        .output)

    print(next(s1)['title'], next(s2)['count'])



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipestrfind
~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for searching for a portion of a string.

You end a `start` and/or `end` string to search for, and a string to search.
If your input string is "ABCDEFG", then a `start` string of `B` gives you a
resulting string of "CDEFG". You can optionally search with regex patterns as
well.

Examples:
    basic usage::

        >>> from riko.modules.pipestrfind import pipe
        >>> conf = {'start': 'hel'}
        >>> next(pipe({'content': 'hello world'}, conf=conf))['strfind']
        u'lo world'

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import re

from builtins import *

from . import processor
from riko.lib.log import Logger

OPTS = {'ftype': 'text', 'ptype': 'text', 'field': 'content'}
DEFAULTS = {'regex': False}
logger = Logger(__name__).logger


def parser(word, objconf, skip, **kwargs):
    """ Parses the pipe content

    Args:
        word (str): The string to parse
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: strfind)
        stream (dict): The original item

    Returns:
        Tuple(dict, bool): Tuple of (item, skip)

    Examples:
        >>> from riko.lib.utils import Objectify
        >>>
        >>> item = {'content': 'hello world'}
        >>> conf = {'start': 'hel'}
        >>> kwargs = {'stream': item, 'conf': conf}
        >>> parser(item['content'], Objectify(conf), False, **kwargs)[0]
        u'lo world'
    """
    if skip:
        value = kwargs['stream']
    elif objconf.regex:
        s = re.split(objconf.start, word)[1] if objconf.start else word
        value = re.split(objconf.end, s)[0] if objconf.end else s
    else:
        s = word.split(objconf.start)[1] if objconf.start else word
        value = s.split(objconf.end)[0] if objconf.end else s

    return value.strip(), skip


@processor(DEFAULTS, async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A processor module that asynchronously finds a portion of a field string
    in an item.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain the keys 'start',
            'end', or 'regex'.

            start (text): starting string (returns everything after this)
            end (text): ending string (returns everything before this)
            regex (bool): interpret `start` and `end` as regex patterns (
                default: False)

        assign (str): Attribute to assign parsed content (default: strfind)
        field (str): Item attribute from which to obtain the first number to
            operate on (default: 'content')

    Returns:
        Deferred: twisted.internet.defer.Deferred item with the found string

    Examples:
        >>> from twisted.internet.task import react
        >>> from riko.twisted import utils as tu
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['strfind'])
        ...     conf = {'start': 'hel'}
        ...     d = asyncPipe({'content': 'hello world'}, conf=conf)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        lo world
    """
    return parser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A processor that finds a portion of a field string in an item.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain the keys 'start',
            'end', or 'regex'.

            start (text): starting string (returns everything after this)
            end (text): ending string (returns everything before this)
            regex (bool): interpret `start` and `end` as regex patterns (
                default: False)

        assign (str): Attribute to assign parsed content (default: strfind)
        field (str): Item attribute from which to obtain the first number to
            operate on (default: 'content')

    Yields:
        dict: an item with the found string

    Examples:
        >>> conf = {'start': 'hel'}
        >>> next(pipe({'content': 'hello world'}, conf=conf))['strfind']
        u'lo world'
        >>> conf = {'start': 'r.*t', 'regex': 'true'}
        >>> next(pipe({'content': 'Greetings'}, conf=conf))['strfind']
        u'ings'
        >>> conf = {'start': 'r', 'end': 'gs'}
        >>> kwargs = {'conf': conf, 'field': 'title', 'assign': 'result'}
        >>> next(pipe({'title': 'Greetings'}, **kwargs))['result']
        u'eetin'
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipefetchdata
~~~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for fetching XML and JSON data sources.

Accesses and extracts data from XML and JSON data sources on the web. This data
can then be converted into an RSS feed or merged with other data in your Pipe.

Examples:
    basic usage::

        >>> from . import FILES
        >>> from riko.modules.pipefetchdata import pipe
        >>>
        >>> conf = {'url': FILES[2], 'path': 'value.items'}
        >>> next(pipe(conf=conf))['title']
        u'Business System Analyst'

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from functools import reduce
from json import loads
from os.path import splitext

from builtins import *
from six.moves.urllib.request import urlopen
from lxml import objectify, html
from lxml.html import html5parser
from twisted.internet.defer import inlineCallbacks, returnValue

from . import processor
from riko.lib import utils
from riko.lib.log import Logger
from riko.twisted import utils as tu

OPTS = {'ftype': 'none'}
reducer = lambda element, i: element.get(i) if element else None
json2dict = lambda f: loads(f.read())
logger = Logger(__name__).logger


def xml2dict(f):
    tree = objectify.parse(f)
    root = tree.getroot()
    return utils.etree_to_dict(root)


def html2dict(f, html5=False):
    tree = html5parser.parse(f) if html5 else html.parse(f)
    root = tree.getroot()
    return utils.etree_to_dict(root)


@inlineCallbacks
def asyncParser(_, objconf, skip, **kwargs):
    """ Asynchronously parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>> from riko.lib.utils import Objectify
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(x[0][0]['title'])
        ...     objconf = Objectify({'url': FILES[2], 'path': 'value.items'})
        ...     kwargs = {'stream': {}}
        ...     d = asyncParser(None, objconf, False, **kwargs)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        Business System Analyst
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        ext = splitext(url)[1].lstrip('.')
        path = objconf.path.split('.') if objconf.path else []
        f = yield tu.urlOpen(url)

        if ext == 'xml':
            element = xml2dict(f)
        elif ext == 'json':
            element = json2dict(f)
        elif ext == 'html':
            element = html2dict(f, objconf.html5)
        else:
            raise TypeError('Invalid file type %s' % ext)

        stream = yield tu.coopReduce(reducer, path, element)

    result = (stream, skip)
    returnValue(result)


def parser(_, objconf, skip, **kwargs):
    """ Parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from . import FILES
        >>> from riko.lib.utils import Objectify
        >>>
        >>> objconf = Objectify({'url': FILES[2], 'path': 'value.items'})
        >>> kwargs = {'stream': {}}
        >>> result, skip = parser(None, objconf, False, **kwargs)
        >>> result[0]['title']
        u'Business System Analyst'
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        ext = splitext(url)[1].lstrip('.')
        path = objconf.path.split('.') if objconf.path else []
        f = urlopen(url)

        if ext == 'xml':
            element = xml2dict(f)
        elif ext == 'json':
            element = json2dict(f)
        elif ext == 'html':
            element = html2dict(f, objconf.html5)
        else:
            raise TypeError('Invalid file type %s' % ext)

        stream = reduce(reducer, path, element)

    return stream, skip


@processor(async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A source that asynchronously fetches and parses an XML or JSON file to
    return the entries.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'path' or 'html5'.

            url (str): The web site to fetch
            path (str): The path to extract (default: None, i.e., return entire
                page)

            html5 (bool): Use the HTML5 parser (default: False)

    Returns:
        Deferred: twisted.internet.defer.Deferred stream of items

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['title'])
        ...     path = 'value.items'
        ...     conf = {'url': FILES[2], 'path': path}
        ...     d = asyncPipe(conf=conf)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        Business System Analyst
    """
    return asyncParser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A source that fetches and parses an XML or JSON file to
    return the entries.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'. May
            contain the keys 'path' or 'html5'.

            url (str): The web site to fetch
            path (str): The path to extract (default: None, i.e., return entire
                page)

            html5 (bool): Use the HTML5 parser (default: False)

    Returns:
        dict: an iterator of items

    Examples:
        >>> from . import FILES
        >>>
        >>> path = 'value.items'
        >>> conf = {'url': FILES[2], 'path': path}
        >>> next(pipe(conf=conf))['title']
        u'Business System Analyst'
        >>> path = 'appointment'
        >>> conf = {'url': FILES[3], 'path': path}
        >>> next(pipe(conf=conf))['subject']
        'Bring pizza home'
        >>> conf = {'url': FILES[3], 'path': ''}
        >>> next(pipe(conf=conf))['reminder']
        '15'
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""Twisted utility functions"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

import itertools as it

from os import environ
from sys import executable
from functools import partial
from io import StringIO
from html.entities import entitydefs, name2codepoint

from builtins import *
from zope.interface import implementer
from twisted.internet import defer
from twisted.internet.defer import (
    inlineCallbacks, maybeDeferred, gatherResults, returnValue, Deferred)
from twisted.internet.task import (
    Cooperator, coiterate as _coiterate, cooperate as _cooperate)
from twisted.internet.utils import getProcessOutput
from twisted.internet.interfaces import IReactorCore
from twisted.internet.reactor import callLater
from twisted.protocols.basic import FileSender
from twisted.web.client import getPage, downloadPage
from twisted.web.microdom import EntityReference
from twisted.test.proto_helpers import (
    MemoryReactor, AccumulatingProtocol, StringTransport)

from riko.lib.log import Logger
from riko.lib.utils import _make_content

logger = Logger(__name__).logger

asyncNone = defer.succeed(None)
asyncReturn = partial(defer.succeed)
asyncPartial = lambda f, **kwargs: partial(maybeDeferred, f, **kwargs)

global FAKE_REACTOR
FAKE_REACTOR = False


# http://stackoverflow.com/q/26314586/408556
# http://stackoverflow.com/q/8157197/408556
# http://stackoverflow.com/a/33708936/408556
class FileReader(AccumulatingProtocol):
    def __init__(self, filename, transform=None, delay=0):
        self.f = open(filename, 'rb')
        self.transform = transform
        self.delay = delay
        self.producer = FileSender()

    def cleanup(self, *args):
        self.f.close()
        self.producer.stopProducing()

    def resumeProducing(self):
        chunk = ''
        if self.file:
            chunk = self.file.read(self.CHUNK_SIZE)

        if not chunk:
            self.file = None
            self.consumer.unregisterProducer()

            if self.deferred and self.delay:
                callLater(seconds, self.deferred.callback, self.lastSent)
            elif self.deferred:
                self.deferred.callback(self.lastSent)

            self.deferred = None
            return

    def connectionLost(self, reason):
        logger.debug('connectionLost: %s', reason)
        self.cleanup()

    def connectionMade(self):
        logger.debug('Connection made from %s', self.transport.getPeer())
        args = (self.f, self.transport, self.transform)
        self.d = self.closedDeferred = self.producer.beginFileTransfer(*args)

        while not self.d.called:
            self.producer.resumeProducing()

        self.d.addErrback(logger.error)
        self.d.addBoth(self.cleanup)


class FakeDelayedCall(object):
    """Fake delayed call which lets us simulate the scheduler.
    """
    def __init__(self, func):
        """A function to run, later.
        """
        self.func = func
        self.cancelled = False

    def cancel(self):
        """Don't run my function later.
        """
        self.cancelled = True


@implementer(IReactorCore)
class FakeReactor(MemoryReactor):
    """A fake reactor to be used in tests.  This reactor doesn't actually do
    much that's useful yet.  It accepts TCP connection setup attempts, but
    they will never succeed.

    Examples:
        >>> import sys
        >>> from twisted.internet.abstract import FileDescriptor
        >>> from twisted.internet.fdesc import readFromFD, setNonBlocking
        >>>
        >>> # reactor = proto_helpers.FakeReactor()
        >>> reactor = FakeReactor()
        >>> f = FileDescriptor(reactor)
        >>> f.fileno = sys.__stdout__.fileno
        >>> fd = f.fileno()
        >>> setNonBlocking(fd)
        >>> readFromFD(fd, print)
    """
    def __init__(self):
        super(FakeReactor, self).__init__()
        global FAKE_REACTOR
        FAKE_REACTOR = True
        self.work = []
        self.running = False

    def resolve(self, *args, **kw):
        """Return a L{twisted.internet.defer.Deferred} that will resolve a hostname.
        """
        pass

    def run(self):
        """Fake L{IReactorCore.run}.
        """
        self.running = True

    def stop(self):
        """Fake L{IReactorCore.stop}.
        """
        self.running = False

    def crash(self):
        """Fake L{IReactorCore.crash}.
        """
        self.running = False

    def iterate(self, *args, **kw):
        """Fake L{IReactorCore.iterate}.
        """
        pass

    def fireSystemEvent(self, *args, **kw):
        """Fake L{IReactorCore.fireSystemEvent}.
        """
        pass

    def addSystemEventTrigger(self, *args, **kw):
        """Fake L{IReactorCore.addSystemEventTrigger}.
        """
        pass

    def removeSystemEventTrigger(self, *args, **kw):
        """Fake L{IReactorCore.removeSystemEventTrigger}.
        """
        pass

    def callWhenRunning(self, *args, **kw):
        """Fake L{IReactorCore.callWhenRunning}.
        """
        pass

    def getDelayedCalls(self):
        """Return all the outstanding delayed calls in the system.
        """
        return (x for x in self.work if not x.cancelled)

    def callLater(self, func):
        """Schedule a unit of work to be done later.
        """
        unit = FakeDelayedCall(func)
        self.work = it.chain(self.work, [unit])
        self.pump()
        return unit

    def pump(self):
        """Perform scheduled work
        """
        for unit in self.getDelayedCalls():
            try:
                unit.func()
            except Exception as e:
                logger.error(e)

        return

cooperator = Cooperator(scheduler=FakeReactor().callLater)


def cleanup(*args):
    if FAKE_REACTOR and cooperator._delayedCall:
        cooperator._delayedCall.cancel()
        cooperator._delayedCall = None


def asyncSleep(seconds):
    d = Deferred()
    callLater(seconds, d.callback, None)
    return d


@inlineCallbacks
def readFile(filename, transport, protocol=FileReader, **kwargs):
    proto = protocol(filename.replace('file://', ''), **kwargs)
    proto.makeConnection(transport)
    yield proto.d
    # returnValue(proto.data)
    returnValue(proto.transport.value())


@inlineCallbacks
def getFile(filename, transport, protocol=FileReader, **kwargs):
    proto = protocol(filename.replace('file://', ''), **kwargs)
    proto.makeConnection(transport)
    yield proto.d
    proto.transport.io.seek(0)
    returnValue(proto.transport.io)


@inlineCallbacks
def urlOpen(url, timeout=0, **kwargs):
    if url.startswith('http'):
        f = StringIO()
        yield downloadPage(url, f, timeout=timeout)
        f.seek(0)
    else:
        f = yield getFile(url, StringTransport(), **kwargs)

    returnValue(f)


def urlRead(url, timeout=0, **kwargs):
    if url.startswith('http'):
        content = getPage(url, timeout=timeout)
    else:
        content = readFile(url, StringTransport(), **kwargs)

    return content


def deferToProcess(source, function, *args, **kwargs):
    command = "from %s import %s\n%s(*%s, **%s)" % (
        source, function, function, args, kwargs)

    return getProcessOutput(executable, ['-c', command], environ)


@inlineCallbacks
def coopReduce(func, iterable, initializer=None):
    cooperate = cooperator.cooperate if FAKE_REACTOR else _cooperate
    iterable = iter(iterable)
    x = initializer or next(iterable)
    result = {}

    def work(func, it, x):
        for y in it:
            result['value'] = x = func(x, y)
            yield

    task = cooperate(work(func, iterable, x))
    yield task.whenDone()
    cleanup()
    returnValue(result['value'])


def asyncReduce(asyncCallable, iterable, initializer=None):
    it = iter(iterable)
    x = initializer or next(it)

    @inlineCallbacks
    def work(asyncCallable, it, x):
        for y in it:
            x = yield asyncCallable(x, y)

        returnValue(x)

    return work(asyncCallable, it, x)


@inlineCallbacks
def pMap(func, iterable, workers=1):
    """map for synchronous callables using parallel cooperative
    multitasking
    """
    coiterate = cooperator.coiterate if FAKE_REACTOR else _coiterate
    results = []

    def work():
        for x in iterable:
            results.append(func(x))
            yield

    deferreds = it.repeat(coiterate(work()), workers)
    yield gatherResults(deferreds, consumeErrors=True)
    cleanup()
    returnValue(results)


def asyncImap(asyncCallable, *iterables):
    """map for deferred callables
    """
    deferreds = map(asyncCallable, *iterables)
    return gatherResults(deferreds, consumeErrors=True)


def asyncStarMap(asyncCallable, iterable):
    """itertools.starmap for deferred callables
    """
    deferreds = it.starmap(asyncCallable, iterable)
    return gatherResults(deferreds, consumeErrors=True)


def asyncDispatch(split, *asyncCallables, **kwargs):
    return asyncStarMap(lambda item, f: f(item), zip(split, asyncCallables))


def asyncBroadcast(item, *asyncCallables, **kwargs):
    return asyncDispatch(it.repeat(item), *asyncCallables, **kwargs)


def def2unicode(entitydef):
    """Convert an HTML entity reference into unicode.
    """
    def2name = {v: k for k, v in entitydefs.items()}
    name = def2name[entitydef]
    cp = name2codepoint[name]
    return chr(cp)


def elementToDict(element, tag='content'):
    """Convert a microdom element into a dict imitating how Yahoo Pipes does it.

    TODO: checkout twisted.words.xish
    """
    i = dict(element.attributes) if hasattr(element, 'attributes') else {}
    value = element.nodeValue if hasattr(element, 'nodeValue') else None

    if isinstance(element, EntityReference):
        value = def2unicode(value)

    i.update(_make_content(i, value, tag))

    for child in element.childNodes:
        tag = child.tagName if hasattr(child, 'tagName') else 'content'
        value = elementToDict(child, tag)

        # try to join the content first since microdom likes to split up
        # elements that contain a mix of text and entity reference
        try:
            i.update(_make_content(i, value, tag, append=False))
        except TypeError:
            i.update(_make_content(i, value, tag))

    if ('content' in i) and not set(i).difference(['content']):
        # element is leaf node and doesn't have attributes
        i = i['content']

    return i



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.pipefeedautodiscovery
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for finding the all available RSS and Atom feeds in a web
site.

Lets you enter a url and then examines those pages for information (like link
rel tags) about available feeds. If information about more than one feed is
found, then multiple items are returned. Because more than one feed can be
returned, the output from this module is often piped into a Fetch Feed module.

Also note that not all sites provide auto-discovery links on their web site's
home page. For a simpler alternative, try the Fetch Site Feed Module. It
returns the content from the first discovered feed.

Examples:
    basic usage::

        >>> from . import FILES
        >>> from riko.modules.pipefeedautodiscovery import pipe
        >>>
        >>> entry = next(pipe(conf={'url': FILES[4]}))
        >>> sorted(entry.keys()) == ['href', 'hreflang', 'link', 'rel', 'tag']
        True
        >>> entry['link']
        'file://data/www.greenhughes.com_rssfeed.xml'

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
from __future__ import (
    absolute_import, division, print_function, unicode_literals)

from builtins import *
from twisted.internet.defer import inlineCallbacks, returnValue

from . import processor
from riko.lib import utils, autorss
from riko.lib.log import Logger

OPTS = {'ftype': 'none'}
logger = Logger(__name__).logger


@inlineCallbacks
def asyncParser(_, objconf, skip, **kwargs):
    """ Asynchronously parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Deferred Tuple of (stream, skip)

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>> from riko.twisted import utils as tu
        >>> from riko.lib.utils import Objectify
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x[0])['link'])
        ...     objconf = Objectify({'url': FILES[4]})
        ...     kwargs = {'stream': {}}
        ...     d = asyncParser(None, objconf, False, **kwargs)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        file://data/www.greenhughes.com_rssfeed.xml
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        stream = yield autorss.asyncGetRSS(url)

    result = (stream, skip)
    returnValue(result)


def parser(_, objconf, skip, **kwargs):
    """ Parses the pipe content

    Args:
        _ (None): Ignored
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        stream (dict): The original item

    Returns:
        Tuple(Iter[dict], bool): Tuple of (stream, skip)

    Examples:
        >>> from . import FILES
        >>> from riko.lib.utils import Objectify
        >>>
        >>> objconf = Objectify({'url': FILES[4]})
        >>> kwargs = {'stream': {}}
        >>> result, skip = parser(None, objconf, False, **kwargs)
        >>> next(result)['link']
        'file://data/www.greenhughes.com_rssfeed.xml'
    """
    if skip:
        stream = kwargs['stream']
    else:
        url = utils.get_abspath(objconf.url)
        stream = autorss.get_rss(url)

    return stream, skip


@processor(async=True, **OPTS)
def asyncPipe(*args, **kwargs):
    """A source that fetches and parses the first feed found on a site.

    Args:
        item (dict): The entry to process (not used)
        kwargs (dict): The keyword arguments passed to the wrapper.

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'.

            url (str): The web site to fetch

    Returns:
        dict: twisted.internet.defer.Deferred an iterator of items

    Examples:
        >>> from twisted.internet.task import react
        >>> from . import FILES
        >>> from riko.twisted import utils as tu
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['link'])
        ...     d = asyncPipe(conf={'url': FILES[4]})
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=tu.FakeReactor())
        ...     pass
        ... except SystemExit:
        ...     pass
        ...
        file://data/www.greenhughes.com_rssfeed.xml
    """
    return asyncParser(*args, **kwargs)


@processor(**OPTS)
def pipe(*args, **kwargs):
    """A source that fetches and parses the first feed found on a site.

    Args:
        item (dict): The entry to process (not used)
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. Must contain the key 'url'.

            url (str): The web site to fetch

    Yields:
        dict: item

    Examples:
        >>> from . import FILES
        >>> next(pipe(conf={'url': FILES[4]}))['link']
        'file://data/www.greenhughes.com_rssfeed.xml'
    """
    return parser(*args, **kwargs)
